"""MIT License

Copyright (c) 2024-2025 gunyu1019

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import asyncio
import aiohttp.web
import webbrowser

from typing import Optional, TYPE_CHECKING
from yarl import URL

if TYPE_CHECKING:
    from .authorization import AccessToken
    from .client import Client


class ChzzkOAuth2Client:
    def __init__(
        self,
        client: Client,
        remote_scheme: Optional[str] = "http",
        remote_host: str = "localhost",
        remote_port: int = 8080,
        remote_path: Optional[str] = "/",
        success_response: Optional[aiohttp.web.Resource] = None,
    ):
        self.client = client

        self.remote_scheme = remote_scheme
        self.remote_host = remote_host
        self.remote_path = remote_path
        self.remote_port = remote_port

        self.success_response = success_response or aiohttp.web.Response(
            text="Success to login!"
        )

        self._application = aiohttp.web.Application()
        self._runner = aiohttp.web.AppRunner(self._application)
        self._site: Optional[aiohttp.web.TCPSite] = None

        self._application.router.add_get(self.remote_path, handler=self.handle)

        self._authorize_state: Optional[str] = None
        self._response: Optional[AccessToken] = None
        self._stop_event = asyncio.Event()

    async def handle(self, request: aiohttp.web.Request):
        code = request.query.get("code")
        state = request.query.get("state")

        if state is None or code is None:
            return aiohttp.web.Response(
                text="Missing requirement parameter. Please try again!"
            )

        if state != self._authorize_state:
            return aiohttp.web.Response(text="Invalid state code. Please try again!")

        self._response = await self.client.generate_access_token(
            code, state=self._authorize_state
        )
        self._stop_event.set()
        return self.success_response

    def _open_webbrowser(self, url: str) -> bool:
        return webbrowser.open(url)

    @property
    def is_server_run(self) -> bool:
        return self._site is not None

    async def close(self):
        if not self.is_server_run:
            raise RuntimeError("A local web server is not enabled.")

        await self._site.stop()
        await self._runner.cleanup()

        self._stop_event.clear()

    async def process_oauth2(
        self, state: str, redirect_url: Optional[str] = None
    ) -> AccessToken:
        if self.is_server_run:
            raise RuntimeError("The OAuth2 login process is already in progress.")

        self._authorize_state = state

        redirect_url = (
            redirect_url
            or URL.build(
                scheme=self.remote_scheme,
                host=self.remote_host,
                path=self.remote_path,
                port=self.remote_port,
            ).__str__()
        )
        url = self.client.generate_authorization_token_url(
            redirect_url=redirect_url, state=state
        )

        await self._runner.setup()

        self._site = aiohttp.web.TCPSite(
            self._runner, self.remote_host, self.remote_port
        )
        await self._site.start()

        open_result = self._open_webbrowser(url)
        if not open_result:
            print("Failed to open the browser. Please join this url and process login.")
        else:
            print(
                "If you haven't opened your browser, please join this url and process login."
            )
        print(url)
        await self._stop_event.wait()

        await self.close()
        return self._response



"""MIT License

Copyright (c) 2024-2025 gunyu1019

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import asyncio
import aiohttp
import logging
from typing import Annotated, Optional, Literal, overload

from ahttp_client import (
    Session,
    request,
    get,
    post,
    put,
    delete,
    BodyJson,
    Query,
    Header,
    Path,
)
from ahttp_client.extension import pydantic_response_model
from ahttp_client.request import RequestCore

from .authorization import AccessToken
from .base_model import Content, SearchResult, ChannelSearchResult
from .category import CATEGORY_TYPE, Category
from .channel import Channel, ChannelPermission, FollowerInfo, SubscriberInfo
from .chat import ChatSetting
from .error import (
    LoginRequired,
    BadRequestException,
    UnauthorizedException,
    ForbiddenException,
    NotFoundException,
    TooManyRequestsException,
    HTTPException,
)
from .live import BrodecastSetting, Live
from .restriction import RestrictUser
from .session import SessionKey

_log = logging.getLogger(__name__)


class ChzzkOpenAPISession(Session):
    def __init__(
        self,
        client_id: str,
        client_secret: str,
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ):
        self.client_id = client_id
        self.client_secret = client_secret
        super().__init__(base_url="https://openapi.chzzk.naver.com", loop=loop)

    async def before_request(
        self, request: RequestCore, path: str
    ) -> tuple[RequestCore, str]:
        _log.debug(f"Path({path}) was called.")

        if hasattr(request.func, "__authorization_configuration__"):
            authorization_configuration = request.func.__authorization_configuration__

            if authorization_configuration["user"]:
                for key, value in request.headers.items():
                    if not isinstance(value, AccessToken):
                        continue

                    access_token = request.headers.pop(key)
                    request.headers["Authorization"] = (
                        f"{access_token.token_type} {access_token.access_token}"
                    )
                    break
                else:
                    if not authorization_configuration["client"]:
                        raise LoginRequired()

            if authorization_configuration["client"]:
                request.headers["Client-Id"] = self.client_id
                request.headers["Client-Secret"] = self.client_secret

        request.headers["Content-Type"] = "application/json"
        return request, path

    async def after_request(self, response: aiohttp.ClientResponse):
        if response.status == 400:
            raise BadRequestException(response.reason)
        elif response.status == 401:
            data = await response.json()
            raise UnauthorizedException(data.get("message"))
        elif response.status == 403:
            data = await response.json()
            raise ForbiddenException(data.get("message"))
        elif response.status == 404:
            data = await response.json()
            raise NotFoundException(data.get("message"))
        elif response.status == 429:
            raise TooManyRequestsException(response.reason)
        elif response.status >= 400:
            data = await response.json()
            raise HTTPException(code=data["code"], message=data["message"])
        return response

    @staticmethod
    async def query_to_json(session: Session, request: RequestCore, path: str):
        copied_request_obj = request.copy()
        body = dict()
        for key, value in request.params.copy().items():
            body[key] = value
        copied_request_obj.params = dict()
        copied_request_obj.body = body
        return copied_request_obj, path

    @staticmethod
    def authorization_configuration(is_client: bool = False, is_user: bool = False):
        def decorator(func):
            func.__authorization_configuration__ = {
                "client": is_client,
                "user": is_user,
            }
            return func

        return decorator

    @overload
    async def generate_access_token(
        self,
        grant_type: Annotated[Literal["authorization_code"], BodyJson.to_camel()],
        client_id: Annotated[str, BodyJson.to_camel()],
        client_secret: Annotated[str, BodyJson.to_camel()],
        code: Annotated[Optional[str], BodyJson.to_camel()],
        state: Annotated[Optional[str], BodyJson.to_camel()],
    ) -> Content[AccessToken]:
        pass

    @overload
    async def generate_access_token(
        self,
        grant_type: Annotated[Literal["refresh_token"], BodyJson.to_camel()],
        client_id: Annotated[str, BodyJson.to_camel()],
        client_secret: Annotated[str, BodyJson.to_camel()],
        refresh_token: Annotated[Optional[str], BodyJson.to_camel()],
    ) -> Content[AccessToken]:
        pass

    @pydantic_response_model()
    @post("/auth/v1/token", directly_response=True)
    @authorization_configuration(is_client=True, is_user=False)
    async def generate_access_token(
        self,
        grant_type: Annotated[
            Literal["authorization_code", "refresh_token"], BodyJson.to_camel()
        ],
        client_id: Annotated[str, BodyJson.to_camel()],
        client_secret: Annotated[str, BodyJson.to_camel()],
        code: Annotated[Optional[str], BodyJson.to_camel()] = None,
        state: Annotated[Optional[str], BodyJson.to_camel()] = None,
        refresh_token: Annotated[Optional[str], BodyJson.to_camel()] = None,
    ) -> Content[AccessToken]:
        pass

    @post("/auth/v1/token/revoke", directly_response=True)
    @authorization_configuration(is_client=True, is_user=False)
    async def revoke_access_token(
        self,
        client_id: Annotated[str, BodyJson.to_camel()],
        client_secret: Annotated[str, BodyJson.to_camel()],
        token: Annotated[Optional[str], BodyJson.to_camel()],
        token_type_hint: Annotated[
            Literal["access_token", "refresh_token"], BodyJson.to_camel()
        ] = "access_token",
    ) -> None:
        pass

    @pydantic_response_model()
    @get("/open/v1/channels", directly_response=True)
    @authorization_configuration(is_client=True, is_user=False)
    async def get_channel(
        self, channel_ids: Annotated[str, Query.to_camel()]
    ) -> Content[SearchResult[Channel]]:
        pass

    @pydantic_response_model()
    @get("/open/v1/channels/streaming-roles", directly_response=True)
    @authorization_configuration(is_client=False, is_user=True)
    async def get_channel_administrator(
        self, token: Annotated[Optional[AccessToken], Header] = None
    ) -> Content[list[ChannelPermission]]:
        pass

    @pydantic_response_model()
    @get("/open/v1/channels/followers", directly_response=True)
    @authorization_configuration(is_client=False, is_user=True)
    async def get_channel_followers(
        self,
        token: Annotated[Optional[AccessToken], Header] = None,
        page: Annotated[int, Query] = 0,
        size: Annotated[int, Query] = 30,
    ) -> Content[ChannelSearchResult[FollowerInfo]]:
        pass

    @pydantic_response_model()
    @get("/open/v1/channels/subscribers", directly_response=True)
    @authorization_configuration(is_client=False, is_user=True)
    async def get_channel_subscribers(
        self,
        token: Annotated[Optional[AccessToken], Header] = None,
        page: Annotated[int, Query] = 0,
        size: Annotated[int, Query] = 30,
        sort: Annotated[Literal["RECENT", "LONGER"], Query] = "RECENT",
    ) -> Content[ChannelSearchResult[SubscriberInfo]]:
        pass

    @pydantic_response_model()
    @get("/open/v1/categories/search", directly_response=True)
    @authorization_configuration(is_client=True, is_user=False)
    async def get_category(
        self, query: Annotated[str, Query], size: Annotated[Optional[int], Query] = 20
    ) -> Content[SearchResult[Category]]:
        pass

    @pydantic_response_model()
    @get("/open/v1/sessions/auth/client", directly_response=True)
    @authorization_configuration(is_client=True, is_user=False)
    async def generate_client_session(self) -> Content[SessionKey]:
        pass

    @pydantic_response_model()
    @get("/open/v1/sessions/auth", directly_response=True)
    @authorization_configuration(is_client=False, is_user=True)
    async def generate_user_session(
        self, token: Annotated[Optional[AccessToken], Header] = None
    ) -> Content[SessionKey]:
        pass

    @pydantic_response_model()
    @post("/open/v1/sessions/events/subscribe/{event}", directly_response=True)
    @authorization_configuration(is_client=True, is_user=True)
    async def subcribe_event(
        self,
        event: Annotated[str, Path],
        session_key: Annotated[str, Query.to_camel()],
        token: Annotated[Optional[AccessToken], Header] = None,
    ) -> Content[None]:
        pass

    @pydantic_response_model()
    @post("/open/v1/sessions/events/unsubscribe/{event}", directly_response=True)
    @authorization_configuration(is_client=True, is_user=True)
    async def unsubcribe_event(
        self,
        event: Annotated[str, Path],
        session_key: Annotated[str, Query.to_camel()],
        token: Annotated[Optional[AccessToken], Header] = None,
    ) -> Content[None]:
        pass

    @pydantic_response_model()
    @get("/open/v1/users/me", directly_response=True)
    @authorization_configuration(is_client=False, is_user=True)
    async def get_user_self(
        self,
        token: Annotated[AccessToken, Header],
    ) -> Content[Channel]:
        pass

    @pydantic_response_model()
    @post("/open/v1/chats/send", directly_response=True)
    @authorization_configuration(is_client=False, is_user=True)
    async def create_message(
        self, token: Annotated[AccessToken, Header], message: Annotated[str, BodyJson]
    ) -> Content[dict[str, str]]:
        pass

    @post("/open/v1/chats/notice", directly_response=True)
    @authorization_configuration(is_client=False, is_user=True)
    async def create_notice(
        self,
        token: Annotated[AccessToken, Header],
        message: Annotated[Optional[str], BodyJson] = None,
        message_id: Annotated[Optional[str], BodyJson.to_camel()] = None,
    ) -> None:
        pass

    @pydantic_response_model()
    @get("/open/v1/chats/settings", directly_response=True)
    @authorization_configuration(is_client=False, is_user=True)
    async def get_chat_setting(
        self,
        token: Annotated[AccessToken, Header],
    ) -> Content[ChatSetting]:
        pass

    @put("/open/v1/chats/settings", directly_response=True)
    @authorization_configuration(is_client=False, is_user=True)
    async def set_chat_setting(
        self,
        token: Annotated[AccessToken, Header],
        chat_available_condition: Annotated[
            Literal["NONE", "REAL_NAME"], BodyJson.to_camel()
        ],
        chat_available_group: Annotated[
            Literal["ALL", "FOLLOWER", "MANAGER", "SUBSCRIBER"], BodyJson.to_camel()
        ],
        min_follower_minute: Annotated[int, BodyJson.to_camel()],
        allow_subscriber_in_follower_mode: Annotated[bool, BodyJson.to_camel()],
        chat_slow_mode_sec: Annotated[
            Literal[0, 3, 5, 10, 30, 60, 120, 300], BodyJson.to_camel()
        ],
        chat_emoji_mode: Annotated[bool, BodyJson.to_camel()],
    ) -> None:
        pass

    @pydantic_response_model()
    @get("/open/v1/lives/setting", directly_response=True)
    @authorization_configuration(is_client=False, is_user=True)
    async def get_live_setting(
        self, token: Annotated[AccessToken, Header]
    ) -> Content[BrodecastSetting]:
        pass

    @request("patch", "/open/v1/lives/setting", directly_response=True)
    @authorization_configuration(is_client=False, is_user=True)
    async def set_live_setting(
        self,
        token: Annotated[AccessToken, Header],
        default_live_title: Annotated[Optional[str], BodyJson.to_camel()] = None,
        category_type: Annotated[Optional[CATEGORY_TYPE], BodyJson.to_camel()] = None,
        category_id: Annotated[Optional[str], BodyJson.to_camel()] = None,
        tags: Annotated[Optional[list[str]], BodyJson.to_camel()] = None,
    ) -> None:
        pass

    @pydantic_response_model()
    @get("/open/v1/streams/key", directly_response=True)
    @authorization_configuration(is_client=False, is_user=True)
    async def get_stream_key(
        self, token: Annotated[AccessToken, Header]
    ) -> Content[dict[str, str]]:
        pass

    @pydantic_response_model()
    @get("/open/v1/lives", directly_response=True)
    @authorization_configuration(is_client=True, is_user=False)
    async def get_lives(
        self,
        size: Annotated[Optional[int], Query] = 20,
        next: Annotated[Optional[str], Query] = None,
    ) -> Content[SearchResult[Live]]:
        pass

    @post("/open/v1/restrict-channels", directly_response=True)
    @authorization_configuration(is_client=False, is_user=True)
    async def add_restrcit_user(
        self,
        token: Annotated[AccessToken, Header],
        target_channel_id: Annotated[str, BodyJson.to_camel()],
    ) -> None:
        pass

    @delete("/open/v1/restrict-channels", directly_response=True)
    @authorization_configuration(is_client=False, is_user=True)
    async def remove_restrcit_user(
        self,
        token: Annotated[AccessToken, Header],
        target_channel_id: Annotated[str, BodyJson.to_camel()],
    ) -> None:
        pass

    @pydantic_response_model()
    @get("/open/v1/restrict-channels", directly_response=True)
    @authorization_configuration(is_client=False, is_user=True)
    async def get_restrcit_users(
        self,
        token: Annotated[AccessToken, Header],
        size: Annotated[Optional[int], Query] = 20,
        next: Annotated[Optional[str], Query] = None,
    ) -> Content[SearchResult[RestrictUser]]:
        pass



"""MIT License

Copyright (c) 2024-2025 gunyu1019

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import json

from typing import Any, Optional

from .enums import EnginePacketType, SocketPacketType, get_enum


class Packet:
    """Implement integrated packet(socket.io + engine.io)
    Consider the environment, the attachments does not implement and binrary does not support.
    """

    def __init__(
        self,
        engine_packet_type: Optional[EnginePacketType] = None,
        socket_packet_type: Optional[SocketPacketType] = None,
        data: Optional[Any] = None,
        *,
        packet_id: Optional[int] = None,
        namespace: Optional[str] = None,
    ):
        self.engine_packet_type = engine_packet_type or EnginePacketType.NOOP
        self.socket_packet_type = socket_packet_type

        self.id = packet_id
        self.data = data
        self.namespace = namespace

    @classmethod
    def _decode_socket(cls, engine_packet_type, payload, json_serialize=None):
        packet_type = get_enum(SocketPacketType, int(payload[0:1]))
        data = payload[1:]

        # Empty Data
        if len(data) == 0:
            return cls(engine_packet_type, packet_type)

        # Decode Attachment (Binaray)
        attachment_separator = data.find("-")
        if attachment_separator > 0 and data[0:attachment_separator].isdigit():
            # Unused attachment feature in chzzk Session API.
            # attachment_count = data[0:attachment_separator]
            data = data[attachment_separator + 1 :]

        # Decode Namespace
        namespace = None
        if len(data) > 0 and data.startswith("/"):
            namespace_separtor = data.find(",")
            if namespace_separtor == -1:
                namespace = data
                data = str()
            else:
                namespace = data[0:namespace_separtor]
                data = data[namespace_separtor + 1 :]

            query = namespace.find("?")
            if query >= 0:
                namespace = namespace[0:query]

        # Decode Packet ID
        packet_id = None
        if len(data) > 0 and data[0].isdigit():
            packet_id = int(data[0])
            while len(data) > 0 and data[0].isdigit():
                packet_id *= 10
                packet_id += int(data[0])
                data = data[1:]

        if len(data) > 0:
            data = json_serialize(data)
        return cls(
            engine_packet_type,
            packet_type,
            data,
            packet_id=packet_id,
            namespace=namespace,
        )

    @classmethod
    def decode(cls, payload, json_serialize=None):
        json_serialize = json_serialize or json.loads
        packet_type = get_enum(EnginePacketType, int(payload[0]))

        if packet_type == EnginePacketType.MESSAGE:
            return cls._decode_socket(
                packet_type, payload[1:], json_serialize=json_serialize
            )

        try:
            data = json_serialize(payload[1:])
            if isinstance(data, int):
                raise ValueError
        except ValueError:
            data = payload[1:]
        return cls(packet_type, None, data)

    @property
    def is_socket_packet(self) -> bool:
        return (
            self.socket_packet_type is not None
            and self.engine_packet_type == EnginePacketType.MESSAGE
        )

    def encode(self, json_serialize=None) -> str:
        json_serialize = json_serialize or json.dumps
        encoded_packet = str(self.engine_packet_type.value)
        if self.is_socket_packet:
            encoded_packet += str(self.socket_packet_type)
            encoded_data = []

            data = (
                json_serialize(self.data, separators=(",", ":"))
                if self.data is not None
                else None
            )

            if self.namespace is not None:
                encoded_data.append(self.namespace)

            if self.id is not None:
                encoded_data.append(str(self.id) + data if data is not None else "")
            elif data is not None:
                encoded_data.append(data)

            encoded_packet += ",".join(encoded_data)
            return encoded_packet
        if isinstance(self.data, str):
            encoded_packet += self.data
        elif isinstance(self.data, dict) or isinstance(self.data, list):
            encoded_packet += json_serialize(self.data, separators=(",", ":"))
        elif self.data is not None:
            encoded_packet += str(self.data)
        return encoded_packet



"""MIT License

Copyright (c) 2024-2025 gunyu1019

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from typing import Any, Literal, Optional


class ChzzkpyException(Exception):
    pass


class LoginRequired(ChzzkpyException):
    """Exception that’s raised when a method need login."""

    def __init__(self):
        super(LoginRequired, self).__init__(
            "This method(feature) needs to login. Please use `login()` method."
        )


class BadRequestException(ChzzkpyException):
    """Exception that’s raised for when status code 400 occurs."""

    def __init__(self, message: Optional[str] = None):
        if message is None:
            message = "Invalid input value"
        super(BadRequestException, self).__init__(message)


class UnauthorizedException(ChzzkpyException):
    """Exception that’s raised for when status code 401 occurs."""

    def __init__(self, message: Optional[str] = None):
        if message is None:
            message = "Not Found"
        super(UnauthorizedException, self).__init__(message)


class ForbiddenException(ChzzkpyException):
    """Exception that’s raised for when status code 403 occurs."""

    def __init__(self, message: Optional[str] = None):
        if message is None:
            message = "Invalid permission"
        super(ForbiddenException, self).__init__(message)


class NotFoundException(ChzzkpyException):
    """Exception that’s raised for when status code 404 occurs."""

    def __init__(self, message: Optional[str] = None):
        if message is None:
            message = "Not Found"
        super(NotFoundException, self).__init__(message)


class TooManyRequestsException(ChzzkpyException):
    """Exception that’s raised for when status code 429 occurs."""

    def __init__(self, message: Optional[str] = None):
        if message is None:
            message = "Too many requests, try later."
        super(TooManyRequestsException, self).__init__(message)


class HTTPException(ChzzkpyException):
    """Exception that’s raised when an HTTP request operation fails."""

    def __init__(self, code: int, message: Optional[str] = None):
        if message is None:
            message = f"Reponsed error code ({code})"
        else:
            message += f" ({code})"
        super(HTTPException, self).__init__(message)


class ChatConnectFailed(ChzzkpyException):
    def __init__(self, message: str):
        super(ChatConnectFailed, self).__init__(message)

    @classmethod
    def websocket_upgrade_failed(cls):
        return cls("WebSocket upgrade failed: no PONG packet")

    @classmethod
    def polling_connect_failed(cls, status: int):
        return cls(f"Unexpected status code {status} in connect polling")

    @classmethod
    def max_connection(cls):
        return cls("Maximum number of sessions allowed to connect has been exceeded.")


class ReceiveErrorPacket(ChzzkpyException):
    def __init__(self, transport: Literal["polling", "websocket"], data: Any):
        super().__init__(
            f"{transport} type of connection received packet with an error. "
            f" (data: {self.code})"
        )

        self.current_transport = transport
        self.data = data



"""MIT License

Copyright (c) 2024-2025 gunyu1019

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import functools

from urllib.parse import parse_qs
from .packet import Packet


class Payload:
    max_decode_packets = 16

    def __init__(self, packets=None):
        self.packets: list[Packet] = packets or []

    def encode(self, jsonp_index=None):
        encoded_payload = b""
        for packet in self.packets:
            encoded_packet = packet.encode()
            content_length = len(encoded_packet)
            encoded_content_length = b""
            while content_length > 0:
                encoded_content_length = (content_length % 10).to_bytes(
                    1, "big"
                ) + encoded_content_length
                content_length //= 10

            encoded_payload += (
                b"\x00"  # Binrary Key
                + encoded_content_length
                + b"\xff"
                + encoded_packet.encode()
            )
        if jsonp_index is not None:
            encoded_payload = (
                b"___eio["
                + jsonp_index.encode()
                + b"]("
                + encoded_payload.replace(b'"', b'\\"')
                + b");"
            )
        return encoded_payload

    @classmethod
    def decode(cls, encoded_payload: bytes):
        if len(encoded_payload) == 0:
            return

        # JSONP POST payload starts with 'd='
        if encoded_payload.startswith(b"d="):
            encoded_payload = parse_qs(encoded_payload)[b"d"][0]

        is_binrary = encoded_payload.find(b"\xff") > -1
        packets = []
        index = 0

        if not is_binrary:
            encoded_payload = encoded_payload.decode("utf-8")
            while index < len(encoded_payload):
                content_length, content = encoded_payload[index:].split(":", maxsplit=1)
                packets.append(Packet.decode(content[: int(content_length)]))
                index += int(content_length)
        else:
            while index < len(encoded_payload):
                is_binrary = int(encoded_payload[index])
                index += 1

                raw_content_length, content = encoded_payload[index:].split(
                    b"\xff", maxsplit=1
                )

                index += len(raw_content_length) + 1
                content_length = functools.reduce(
                    lambda d, s: d * 10 + s, raw_content_length
                )

                # Binrary is not supported. (TODO)
                if not is_binrary:
                    content = content[: int(content_length)].decode("utf-8")
                    packets.append(Packet.decode(content))
                index += int(content_length)
        return cls(packets=packets)



"""MIT License

Copyright (c) 2024-2025 gunyu1019

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

from functools import reduce
from typing import (
    Any,
    Callable,
    ClassVar,
    Iterator,
    Type,
    TypeVar,
    Optional,
    overload,
)

BF = TypeVar("BF", bound="BaseFlags")


class flag_value:
    def __init__(self, func: Callable[[Any], int]):
        self.flag: int = func(None)
        self.__doc__: Optional[str] = func.__doc__

    def __get__(self, instance: Optional[BF], owner: Type[BF]) -> Any | bool:
        if instance is None:
            return self
        return instance._has_flag(self.flag)

    def __set__(self, instance: BaseFlags, value: bool) -> None:
        instance._set_flag(self.flag, value)

    def __repr__(self) -> str:
        return f"<flag_value flag={self.flag!r}>"


def fill_with_flags(*, inverted: bool = False) -> Callable[[Type[BF]], Type[BF]]:
    def decorator(cls: Type[BF]) -> Type[BF]:
        cls.VALID_FLAGS = {
            name: value.flag
            for name, value in cls.__dict__.items()
            if isinstance(value, flag_value)
        }

        if inverted:
            max_bits = max(cls.VALID_FLAGS.values()).bit_length()
            cls.DEFAULT_VALUE = -1 + (2**max_bits)
        else:
            cls.DEFAULT_VALUE = 0

        return cls

    return decorator


# n.b. flags must inherit from this and use the decorator above
class BaseFlags:
    VALID_FLAGS: ClassVar[dict[str, int]]
    DEFAULT_VALUE: ClassVar[int]

    value: int

    def __init__(self, **kwargs: bool):
        self.value = self.DEFAULT_VALUE
        for key, value in kwargs.items():
            if key not in self.VALID_FLAGS:
                raise TypeError(f"{key!r} is not a valid flag name.")
            setattr(self, key, value)

    @classmethod
    def _from_value(cls, value: int):
        self = cls.__new__(cls)
        self.value = value
        return self

    def __or__(self, other):
        return self._from_value(self.value | other.value)

    def __and__(self, other):
        return self._from_value(self.value & other.value)

    def __xor__(self, other):
        return self._from_value(self.value ^ other.value)

    def __ior__(self, other):
        self.value |= other.value
        return self

    def __iand__(self, other):
        self.value &= other.value
        return self

    def __ixor__(self, other):
        self.value ^= other.value
        return self

    def __invert__(self):
        max_bits = max(self.VALID_FLAGS.values()).bit_length()
        max_value = -1 + (2**max_bits)
        return self._from_value(self.value ^ max_value)

    def __bool__(self) -> bool:
        return self.value != self.DEFAULT_VALUE

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__) and self.value == other.value

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash(self.value)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} value={self.value}>"

    def __iter__(self) -> Iterator[tuple[str, bool]]:
        for name, value in self.__class__.__dict__.items():
            if isinstance(value, flag_value):
                yield (name, self._has_flag(value.flag))

    def _has_flag(self, o: int) -> bool:
        return (self.value & o) == o

    def _set_flag(self, o: int, toggle: bool) -> None:
        if toggle is True:
            self.value |= o
        elif toggle is False:
            self.value &= ~o
        else:
            raise TypeError(
                f"Value to set for {self.__class__.__name__} must be a bool."
            )


@fill_with_flags()
class UserPermission(BaseFlags):
    """The UserPermission Flag is used by session for event subscriptions."""

    def __init__(self, value: int = 0, **kwargs: bool) -> None:
        self.value: int = value
        for key, value in kwargs.items():
            if key not in self.VALID_FLAGS:
                raise TypeError(f"{key!r} is not a valid flag name.")
            setattr(self, key, value)

    @classmethod
    def all(cls: Type[UserPermission]) -> UserPermission:
        """A factory method that creates a :class:`UserPermission` with everything enabled."""
        value = reduce(lambda a, b: a | b, cls.VALID_FLAGS.values())
        self = cls.__new__(cls)
        self.value = value
        return self

    @flag_value
    def chat(self):
        """:class:`bool`: Whether chat event are enabled."""
        return 1 << 0

    @flag_value
    def donation(self):
        """:class:`bool`: Whether donation event are enabled."""
        return 1 << 1

    @flag_value
    def subscription(self):
        """:class:`bool`: Whether subscription event are enabled."""
        return 1 << 2



"""MIT License

Copyright (c) 2024-2025 gunyu1019

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import datetime

from pydantic import Field, PrivateAttr
from typing import Any, Literal, Optional, TYPE_CHECKING

from .base_model import ChzzkModel

if TYPE_CHECKING:
    from .authorization import AccessToken
    from .state import ConnectionState


class Profile(ChzzkModel):
    nickname: str
    badges: list[Any]
    verified_mark: bool


class Messageable(ChzzkModel):
    _access_token: Optional[AccessToken] = PrivateAttr(default=None)
    _state: Optional[ConnectionState] = PrivateAttr(default=None)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if "state" in kwargs.keys() and "access_token" in kwargs.keys():
            self._access_token = kwargs.pop("access_token")
            self._state = kwargs.pop("state")

    async def send(self, content: str) -> SentMessage:
        """Send a message to the received channel

        Parameters
        ----------
        content : str
            A content of message to send.

        Raises
        ------
        RuntimeError
            Raised when the access token is missing,
            or there is insufficient data to send message due to unusual situations.
        """
        if self._state is None or self._access_token is None:
            raise RuntimeError(
                f"This {self.__class__.__name__} is intended to store data only."
                f"Or don't have the authentication token to send the message."
            )
        message = await self._state.http.create_message(
            message=content, token=self._access_token
        )
        message._state = self._state
        message._access_token = self._access_token
        return message


class Subscription(Messageable):
    """A subscription instance received form the live.
    Subscription can be received via the `on_subscription` event.
    """

    channel: str = Field(alias="channelId")
    subscriber_id: str = Field(alias="subscriberChannelId")
    subscriber_name: str = Field(alias="subscriberNickname")

    tier_no: int
    tier_name: str
    month: int


class Donation(Messageable):
    """A donation instance received form the live.
    Donations can be received via the `on_donation` event.
    """

    type: Literal["CHAT", "VIDEO"] = Field(alias="donationType")
    channel: str = Field(alias="channelId")
    donator_id: str = Field(alias="donatorChannelId")
    donator_name: str = Field(alias="donatorNickname")

    pay_amount: int
    donation_text: str


class Message(Messageable):
    """A message instance received from the live.
    Messages can be received via the `on_chat` event.
    """

    user_id: str = Field(alias="senderChannelId")

    profile: Profile
    content: str
    channel: str = Field(alias="channelId")
    created_time: datetime.datetime = Field(alias="messageTime")


class SentMessage(Messageable):
    """Represents a message sent by client.send() method"""

    id: str
    content: str
    created_time: datetime.datetime = Field(default_factory=datetime.datetime.now)

    async def pin(self) -> None:
        """Pin the sent message for announcement."""
        if self._state is None or self._access_token is None:
            raise RuntimeError(
                f"This {self.__class__.__name__} is intended to store data only."
            )
        await self._state.http.create_notice(
            message_id=self.id, token=self._access_token
        )



"""MIT License

Copyright (c) 2024-2025 gunyu1019

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import asyncio
import aiohttp
import datetime
import logging

from functools import wraps
from typing import Any, overload, TYPE_CHECKING
from yarl import URL

from .authorization import AccessToken
from .chat import ChatSetting
from .error import ChatConnectFailed, ForbiddenException
from .gateway import ChzzkGateway
from .http import ChzzkOpenAPISession
from .live import BrodecastSetting, Live
from .message import SentMessage
from .oauth2 import ChzzkOAuth2Client
from .state import ConnectionState


if TYPE_CHECKING:
    from aiohttp.web import Response as webResponse
    from typing import Self, Literal, Optional, Callable, Coroutine

    from .base_model import SearchResult, ChannelSearchResult
    from .channel import Channel, ChannelPermission, FollowerInfo, SubscriberInfo
    from .category import Category
    from .enums import FollowingPeriod
    from .flags import UserPermission
    from .restriction import RestrictUser


class _LoopSentinel:
    __slots__ = ()

    def __getattr__(self, attr: str) -> None:
        msg = (
            "loop attribute cannot be accessed in non-async contexts. "
            "Consider using either an asynchronous main function and passing it to asyncio.run"
        )
        raise AttributeError(msg)


_log = logging.getLogger(__name__)


class BaseEventManager:
    def __init__(self, loop: asyncio.AbstractEventLoop):
        self.loop = loop
        self._listeners: dict[str, list[tuple[asyncio.Future, Callable[..., bool]]]] = (
            dict()
        )
        self._extra_event: dict[str, list[Callable[..., Coroutine[Any, Any, Any]]]] = (
            dict()
        )

    def wait_for(
        self,
        event: str,
        check: Optional[Callable[..., bool]] = None,
        timeout: Optional[float] = None,
    ):
        """Waits for a WebSocket event to be dispatched.

        Parameters
        ----------
        event : str
            The event name.
            For a list of events, read :method:`event`
        check : Optional[Callable[..., bool]],
            A predicate to check what to wait for. The arguments must meet the
            parameters of the event being waited for.
        timeout : Optional[float]
            The number of seconds to wait before timing out and raising
            :exc:`asyncio.TimeoutError`.

        """
        future = self.loop.create_future()

        if check is None:

            def _check(*_):
                return True

            check = _check
        event_name = event.lower()

        if event_name not in self._listeners.keys():
            self._listeners[event_name] = list()
        self._listeners[event_name].append((future, check))
        return asyncio.wait_for(future, timeout=timeout)

    def event(
        self, coro: Callable[..., Coroutine[Any, Any, Any]]
    ) -> Callable[..., Coroutine[Any, Any, Any]]:
        """A decorator that registers an event to listen to.
        The function must be coroutine. Else client cause TypeError.

        A list of events that the client can listen to.
        * `on_chat`: Called when a Message is created and sent.
        * `on_connect`: Called when the client is done preparing the data received from Chzzk.
        * `on_donation`: Called when a listener donates
        * `on_error`: Called when event raise exception

        Example
        -------
        >>> @client.event
        ... async def on_chat(message: Message):
        ...     print(message.content)
        ...     await message.send("Reply Mesage")
        """
        if not asyncio.iscoroutinefunction(coro):
            raise TypeError("function must be a coroutine.")

        event_name = coro.__name__
        if event_name not in self._listeners.keys():
            self._extra_event[event_name] = list()
        self._extra_event[event_name].append(coro)
        return coro

    def dispatch(self, event: str, *args: Any, **kwargs) -> None:
        _log.debug("Dispatching event %s", event)
        method = "on_" + event

        # wait-for listeners
        if event in self._listeners.keys():
            listeners = self._listeners[event]
            _new_listeners = []

            for index, (future, condition) in enumerate(listeners):
                if future.cancelled():
                    continue

                try:
                    result = condition(*args, **kwargs)
                except Exception as e:
                    future.set_exception(e)
                    continue
                if result:
                    match len(args):
                        case 0:
                            future.set_result(None)
                        case 1:
                            future.set_result(args[0])
                        case _:
                            future.set_result(args)

                _new_listeners.append((future, condition))
            self._listeners[event] = _new_listeners

        # event-listener
        if method not in self._extra_event.keys():
            return

        for coroutine_function in self._extra_event[method]:
            self._schedule_event(coroutine_function, method, *args, **kwargs)

    async def _run_event(
        self,
        coro: Callable[..., Coroutine[Any, Any, Any]],
        event_name: str,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        try:
            await coro(*args, **kwargs)
        except asyncio.CancelledError:
            pass
        except Exception as exc:
            try:
                _log.exception("Ignoring exception in %s", event_name)
                self.dispatch("error", exc, *args, **kwargs)
            except asyncio.CancelledError:
                pass

    def _schedule_event(
        self,
        coro: Callable[..., Coroutine[Any, Any, Any]],
        event_name: str,
        *args: Any,
        **kwargs: Any,
    ) -> asyncio.Task:
        wrapped = self._run_event(coro, event_name, *args, **kwargs)
        # Schedules the task
        return self.loop.create_task(wrapped, name=f"chzzk.py: {event_name}")


class Client(BaseEventManager):
    """Represents a client to connect Chzzk (Naver Live Streaming)."""

    def __init__(
        self,
        client_id: str,
        client_secret: str,
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ):
        super().__init__(loop)
        self.loop = loop or _LoopSentinel()
        self.client_id = client_id
        self.client_secret = client_secret

        self.http: Optional[ChzzkOpenAPISession] = None
        self.user_client: list[UserClient] = []

        handler = {"connect": lambda _: self._gateway_ready.set()}
        self._connection = ConnectionState(
            dispatch=self.dispatch,
            handler=handler,
            http=self.http,
            variable_access_token=self.__variable_access_token,
        )

        self._gateway: dict[str, ChzzkGateway] = dict()
        self._gateway_ready = asyncio.Event()

    def __variable_access_token(self, channel_id: str) -> Optional[AccessToken]:
        user_client = self.get_user_client_cached(channel_id)
        if user_client is not None:
            return user_client.access_token
        return None

    async def __aenter__(self) -> Self:
        await self._async_setup_hook()
        return self

    async def _async_setup_hook(self) -> None:
        if isinstance(self.loop, _LoopSentinel):
            self.loop = asyncio.get_running_loop()

        if asyncio.current_task(self.loop) is None:
            raise RuntimeError(
                "This loop should be defined in an asynchronous context."
            )

        self.http = ChzzkOpenAPISession(
            loop=self.loop, client_id=self.client_id, client_secret=self.client_secret
        )
        self._connection.http = self.http

    @staticmethod
    def initial_async_setup(func):
        @wraps(func)
        async def wrapper(self: Self, *args, **kwargs):
            if isinstance(self.loop, _LoopSentinel) or self.http is None:
                await self._async_setup_hook()
            return await func(self, *args, **kwargs)

        return wrapper

    def generate_authorization_token_url(self, redirect_url: str, state: str) -> str:
        """Get an url to generate chzzk access token.

        Parameters
        ----------
        redirect_url : str
            Redirect URL after login.
        state : str
            A state code to use in `state` parameter of :meth:`generate_access_token` method.
        """
        default_url = URL.build(
            scheme="https", authority="chzzk.naver.com", path="/account-interlock"
        )
        default_url = default_url.with_query(
            {"clientId": self.client_id, "redirectUri": redirect_url, "state": state}
        )
        return default_url.__str__()

    @initial_async_setup
    async def generate_access_token(self, code: str, state: str) -> AccessToken:
        """Generate an Access Token.
        An access token is used by features that require user authentication. (example. creating chat, receiving chat)

        Parameters
        ----------
        code : str
            A code received from :meth:`generate_authorization_token_url`.
        state : str
            A state code.
            This value must be the same as the state of :meth:`generate_authorization_token_url`.

        Returns
        -------
        AccessToken
            A instance that contain user authentic.
        """
        result = await self.http.generate_access_token(
            grant_type="authorization_code",
            client_id=self.client_id,
            client_secret=self.client_secret,
            code=code,
            state=state,
        )
        return result.content

    async def generate_user_client(self, code: str, state: str) -> UserClient:
        """Generate user client instance.

        Parameters
        ----------
        code : str
            A code received from :meth:`generate_authorization_token_url`.
        state : str
            A state code.
            This value must be the same as the state of :meth:`generate_authorization_token_url`.
        """
        access_token = await self.generate_access_token(code, state)
        user_cls = UserClient(self, access_token)
        try:
            await user_cls.fetch_self()
        except ForbiddenException:
            pass
        self.user_client.append(user_cls)
        return user_cls

    @initial_async_setup
    async def login(
        self,
        state: Optional[str] = None,
        redirect_url: Optional[str] = None,
        remote_host: str = "localhost",
        remote_port: int = 8080,
        remote_path: str = "/",
        success_response: Optional[webResponse] = None,
        ssl: bool = False,
    ) -> UserClient:
        """Get :class:`UserClient` instance with oauth2 login process.
        This :meth:`login` provides a convenient login method using the aiohttp web server.

        Parameters
        ----------
        state: Optional[str]
            A code used in the authentication process.
        redirect_url: Optional[str]
            A redirect_url used in the authentication process.
        remote_host : str
            A host url for the temporarily operated web server, default by localhost.
            If the public environment required, enter "0.0.0.0" value to this parameter.
        remote_port : int
            A port for the temporarily operated web server, default by 8080.
        remote_path : int
            A path for the temporarily operated web server, default by '/'.
        success_response : aiohttp.web.Response
            A content to be displayed on the website after successful login.
        ssl: bool
            Whether to use http(s) during opening the web browser.
        """
        remote_scheme = "https" if ssl else "http"
        oauth2_client = ChzzkOAuth2Client(
            self,
            remote_scheme=remote_scheme,
            remote_host=remote_host,
            remote_port=remote_port,
            remote_path=remote_path,
            success_response=success_response,
        )

        state = state or "chzzkpy_authorization"  # Temporatry Code
        access_token = await oauth2_client.process_oauth2(state, redirect_url)
        user_client = await self.get_user_client(access_token)
        return user_client

    @initial_async_setup
    async def refresh_access_token(self, refresh_token: str) -> AccessToken:
        """Regenerate an Access Token.
        The Access Token in chzzk follows Open Authorization 2.0 protocol,
         which means that access tokens expire after a certain amount of time.

        Parameters
        ----------
        refresh_token : str
            A refresh token to re-generated access token.
        """
        refresh_token = await self.http.generate_access_token(
            grant_type="refresh_token",
            client_id=self.client_id,
            client_secret=self.client_secret,
            refresh_token=refresh_token,
        )
        return refresh_token.content

    @initial_async_setup
    async def refresh_user_client(self, refresh_token: str) -> UserClient:
        """Get :class:`UserClient` instance with refresh_token.

        Parameters
        ----------
        refresh_token : str
            A refresh token to re-generated access token.
        """
        refreshed_access_token = await self.refresh_access_token(refresh_token)
        user_cls = UserClient(self, refreshed_access_token)
        try:
            await user_cls.fetch_self()
        except ForbiddenException:
            pass
        self.user_client.append(user_cls)
        return user_cls

    @initial_async_setup
    async def get_user_client(self, access_token: AccessToken) -> UserClient:
        """Generate :class:`UserClient` with prepared access token.

        Parameters
        ----------
        access_token : AccessToken
            A prepared instance of access token.
        """
        user_cls = UserClient(self, access_token)
        try:
            await user_cls.fetch_self()
        except ForbiddenException:
            pass
        self.user_client.append(user_cls)
        return user_cls

    def get_user_client_cached(self, channel_id: str) -> Optional[UserClient]:
        """Get :class:`UserClient` as channel_id.
        The channel id of :class:`UserClient` can be obtained via the subscription event of session
        or when the client has the `유저 정보 조회` scope.

        If the channel Id of :class:`UserClient` is not filled, it can't found client.

        Parameters
        ----------
        channel_id : str
            A channel ID to get :class:`UserClient`
        """
        for user_client in self.user_client:
            if user_client.channel_id != channel_id:
                continue
            return user_client
        return

    @initial_async_setup
    async def get_channel(self, channel_ids: list[str]) -> list[Channel]:
        """Get channel information.

        Parameters
        ----------
        channel_ids : list[str]
            An unique ID of the channel to lookup.
        """
        result = await self.http.get_channel(channel_ids=",".join(channel_ids))
        return result.content.data

    @initial_async_setup
    async def get_category(
        self, query: str, size: Optional[int] = 20
    ) -> list[Category]:
        """Get category information

        Parameters
        ----------
        query : str
            A name of category.
        size : Optional[int], optional
            A number of categories to load at once, by default 20
        """
        result = await self.http.get_category(query=query, size=size)
        return result.content.data

    @initial_async_setup
    async def get_live(self, size: int = 20) -> SearchResult[Live]:
        """Get live information

        Parameters
        ----------
        size : Optional[int], optional
            A number of lives to load at once, by default 20
        """
        result = await self.http.get_lives(size=size)
        data = result.content
        data._next_method = self.http.get_lives
        data._next_method_key_argument = {"size": size}
        return data

    async def wait_until_connect(self):
        """Waits until the session connected."""
        await self._gateway_ready.wait()
        return

    @initial_async_setup
    async def connect(self, addition_connect: bool = False) -> str:
        """Connect to session to handle donation or chatting
        A Client session can have more than one connection. (Maximum Connection: 10)

        Parameters
        ----------
        addition_connect : Optional[bool]
            This parameter used for multiple connections, by default False
            If addition_connect is False, the connection is completed and a main task is blocked to wait for a response.
            However addition_connect is True, a task waiting for a response is processed in the background.

        Returns
        -------
        str
            Returns an unique session ID, if the gateway connects succeeds.
            A session ID used by subscribe, unsubscribe method.

        Warning
        -------
        If the main task is empty,
        as if addition_connect parameter used for all connections is true,
        all connections are aborted.

        Raises
        ------
        ChatConnectFailed.max_connection
            A Client Session can connect up to 10 sessions.
            When connection request more than 10 sessions, ChatConnectFailed exception raised.
        """
        if len(self._gateway.keys()) > 10:
            raise ChatConnectFailed.max_connection()

        session_key = await self.http.generate_client_session()
        gateway_cls = await ChzzkGateway.connect(
            url=session_key.content.url,
            state=self._connection,
            loop=self.loop,
            session=aiohttp.ClientSession(loop=self.loop),
        )
        task = gateway_cls.read_in_background()
        await self._gateway_ready.wait()
        self._gateway[gateway_cls.session_id] = gateway_cls
        self._gateway_ready.clear()
        if not addition_connect:
            await task
        return gateway_cls.session_id

    async def disconnect(self, session_id: Optional[str] = None):
        """Disconnect from session.

        Parameters
        ----------
        session_id : Optional[str]
            A session ID to disconnect from session.
            When session parameter is empty, all client sessions are disconnected.
        """
        if len(self._gateway) <= 0:
            return

        if session_id is not None:
            await self._gateway[session_id].disconnect()
            self._gateway.pop(session_id)
            return

        for gateway in self._gateway.values():
            await gateway.disconnect()
        self._gateway = dict()


class UserClient:
    """Represents a user client to provide feature that requires user authentication."""

    def __init__(self, parent: Client, access_token: AccessToken):
        self.parent_client = parent
        self.dispatch = self.parent_client.dispatch
        self.loop = self.parent_client.loop
        self.http = self.parent_client.http

        self.access_token = access_token
        self._token_generated_at = datetime.datetime.now()

        # Resolved issue in https://github.com/gunyu1019/chzzkpy/issues/66
        if not isinstance(self.access_token, AccessToken):
            raise TypeError(
                "An invalid type of parameter was entered. The access_token parameter must be AccessToken type."
            )

        self._gateway: Optional[ChzzkGateway] = None
        self._gateway_ready = asyncio.Event()
        self._gateway_id: Optional[str] = None
        self._session_id: Optional[str] = None

        self.channel_id: Optional[str] = None
        self.channel_name: Optional[str] = None

        handler = {
            "connect": self.__on_connected,
            "channel_id_invoked": self.__on_channel_id_invoked,
        }
        self._connection = ConnectionState(
            dispatch=self.dispatch,
            handler=handler,
            http=self.http,
            access_token=self.access_token,
        )

    def __on_connected(self, session_id: str):
        self._session_id = session_id
        self._gateway_ready.set()
        return

    def __on_channel_id_invoked(self, channel_id: str):
        self.channel_id = channel_id

    @property
    def is_expired(self) -> bool:
        """An access token for user expires after a certain amount of time.
        Returns status that an access token had been expired."""
        return (
            datetime.datetime.now() - self._token_generated_at
        ).hours > self.access_token.expires_in

    @staticmethod
    def refreshable(func):
        @wraps(func)
        async def wrapper(self: Self, *args, **kwargs):
            remaining_expired_time = datetime.datetime.now() - self._token_generated_at
            if remaining_expired_time.days > 0:
                await self.refresh()
            return await func(self, *args, **kwargs)

        return wrapper

    async def refresh(self):
        """Refresh the access token."""
        refresh_token = await self.http.generate_access_token(
            grant_type="refresh_token",
            client_id=self.parent_client.client_id,
            client_secret=self.parent_client.client_secret,
            refresh_token=self.access_token.refresh_token,
        )
        self._connection.access_token = self.access_token = refresh_token.content
        self._token_generated_at = datetime.datetime.now()
        return

    async def revoke(self):
        """Revoke the access token."""
        await self.http.revoke_access_token(
            client_id=self.parent_client.client_id,
            client_secret=self.parent_client.client_secret,
            token=self.access_token.access_token,
        )
        return

    @refreshable
    async def fetch_self(self) -> Channel:
        """Get channel based on access token.
        This method required "유저 정보 조회" API Scope on the access token.

        Returns
        -------
        Channel
            A channel based on the access token.
        """
        raw_user_self = await self.http.get_user_self(token=self.access_token)
        user_self = raw_user_self.content
        self.channel_id = user_self.id
        self.channel_name = user_self.name
        return user_self

    @refreshable
    async def send_message(self, content: str) -> SentMessage:
        """Send the message to channel
        This method required "채팅 메시지 전송" API Scope on the access token.

        Parameters
        ----------
        content : str
            A content of message.
        """
        response = await self.http.create_message(
            token=self.access_token, message=content
        )
        message_id = response.content["messageId"]
        message = SentMessage(id=message_id, content=content)
        message._access_token = self.access_token
        message._state = self._connection
        return message

    @refreshable
    async def send_announcement(self, content: str):
        """Send the announcement to channel
        This method required "채팅 공지 등록" API Scope on the access token.

        Parameters
        ----------
        content : str
            A content of announcement.
        """
        await self.http.create_notice(token=self.access_token, message=content)
        return

    @property
    def is_connected(self) -> bool:
        """Returns the gateway had been connected."""
        if self._gateway is None:
            return False
        return self._gateway.is_connected

    async def wait_until_connect(self):
        """Waits until the session connected."""
        await self._gateway_ready.wait()
        return

    @refreshable
    async def connect(self, permission: UserPermission, addition_connect: bool = False):
        """Connect to user session to handle donation or chatting

        Parameters
        ----------
        permission: UserPermission
            The permissions to receive, such as chat or donation.
            After connecting, permission is granted via subscription method.
        addition_connect : Optional[bool]
            This parameter used for multiple connections, by default False
            If addition_connect is False, the connection is completed and a main task is blocked to wait for a response.
            However addition_connect is True, a task waiting for a response is processed in the background.

        Returns
        -------
        str
            Returns an unique session ID, if the gateway connects succeeds.
            A session ID used by subscribe, unsubscribe method.

        Warning
        -------
        If the main task is empty,
        as if addition_connect parameter used for all connections is true,
        all connections are aborted.
        """
        session_key = await self.http.generate_user_session(token=self.access_token)
        self._gateway = await ChzzkGateway.connect(
            url=session_key.content.url,
            state=self._connection,
            loop=self.loop,
            session=aiohttp.ClientSession(loop=self.loop),
        )
        task = self._gateway.read_in_background()
        await self._gateway_ready.wait()
        self._gateway_id = self._gateway.session_id
        await self.subscribe(permission, self._session_id)
        if not addition_connect:
            await task
        return self._session_id

    async def disconnect(self):
        """Disconnect from session."""
        if self._gateway is None:
            return
        await self._gateway.disconnect()

        self._gateway = None
        self._gateway_id = None
        self._session_id = None
        self._gateway_ready.clear()

    @refreshable
    async def subscribe(
        self, permission: UserPermission, session_id: Optional[str] = None
    ):
        """Subcribe to events based on UserPermission.
        Subsribed events are recived in the session.

        Parameters
        ----------
        permission : UserPermission
            The session subscribed evnet.
        session_id : Optional[str]
            ID of the session to receive event, by default None
            The session ID is provided as a parameter in `on_connect` event
        """
        session_id = session_id or self._session_id
        if session_id is None:
            raise TypeError(
                "A session_id is not filled. Connect to session using UserClient.connect() method or Client.connect()."
            )

        for permission_name, condition in permission:
            if not condition:
                continue
            await self.http.subcribe_event(
                event=permission_name, session_key=session_id, token=self.access_token
            )
            _log.debug(f"Subscribe {permission_name.upper()} Event")
        return

    @refreshable
    async def unsubscribe(
        self, permission: UserPermission, session_id: Optional[str] = None
    ):
        """Unsubcribe to events based on UserPermission.
        The events specified by UserPermission are no longer received.

        Parameters
        ----------
        permission : UserPermission
            The session unsubscribed evnet.
        session_id : Optional[str]
            ID of the session to block event, by default None
        """
        session_id = session_id or self._gateway_id
        if session_id is None:
            raise TypeError(
                "A session_id is not filled. Connect to session using UserClient.connect() method or Client.connect()."
            )

        for permission_name, condition in permission:
            if not condition:
                continue

            await self.http.unsubcribe_event(
                event=permission_name, session_key=session_id, token=self.access_token
            )
            _log.debug(f"Unsubscribe {permission_name.upper()} Event")
        return

    @refreshable
    async def get_chat_setting(self) -> ChatSetting:
        """Get the chat settings."""
        raw_chat_setting = await self.http.get_chat_setting(token=self.access_token)
        return raw_chat_setting.content

    @overload
    async def set_chat_setting(self, instance: ChatSetting) -> None:
        """Set the chat settings.

        Parameters
        ----------
        instance : ChatSetting
            A instance of chat setting.
        """
        ...

    @overload
    async def set_chat_setting(
        self,
        chat_available_condition: Literal["NONE", "REAL_NAME"],
        chat_available_group: Literal["ALL", "FOLLOWER", "MANAGER", "SUBSCRIBER"],
        min_follower_minute: FollowingPeriod,
        allow_subscriber_in_follower_mode: bool,
        slow_mode: Literal[0, 3, 5, 10, 30, 60, 120, 300] = 0,
        emoji_mode: bool = False,
    ) -> None:
        """Set the chat settings.

        Parameters
        ----------
        chat_available_condition : Literal['NONE', 'REAL_NAME']
            Allow only users who have been authenticated.
        chat_available_group : Literal['ALL', 'FOLLOWER', 'MANAGER', 'SUBSCRIBER']
            Set the types of chat that are allowed.
        min_follower_minute : FollowingPeriod
            Allow users who follow channel to make chat available after a specified amount of time.
        allow_subscriber_in_follower_mode : bool
            Set whether subscriber will follow the minimum follow time.
        slow_mode : Literal[0, 3, 5, 10, 30, 60, 120, 300]
            Set user chat interval (seconds)
        emoji_mode : bool
            Apply emoji mode
        """
        ...

    @refreshable
    async def set_chat_setting(
        self,
        instance: ChatSetting = None,
        chat_available_condition: Literal["NONE", "REAL_NAME"] = None,
        chat_available_group: Literal[
            "ALL", "FOLLOWER", "MANAGER", "SUBSCRIBER"
        ] = None,
        min_follower_minute: FollowingPeriod = None,
        allow_subscriber_in_follower_mode: bool = None,
        slow_mode: Literal[0, 3, 5, 10, 30, 60, 120, 300] = 0,
        emoji_mode: bool = False,
    ) -> None:
        if instance is not None:
            await self.http.set_chat_setting(
                token=self.access_token,
                chat_available_condition=instance.chat_available_condition,
                chat_available_group=instance.chat_available_group,
                min_follower_minute=instance.min_follower_minute,
                allow_subscriber_in_follower_mode=instance.allow_subscriber_in_follower_mode,
            )
            return
        await self.http.set_chat_setting(
            token=self.access_token,
            chat_available_condition=chat_available_condition,
            chat_available_group=chat_available_group,
            min_follower_minute=min_follower_minute,
            allow_subscriber_in_follower_mode=allow_subscriber_in_follower_mode,
            chat_slow_mode_sec=slow_mode,
            chat_emoji_mode=emoji_mode,
        )
        return

    @refreshable
    async def get_live_setting(self) -> BrodecastSetting:
        """Get the live settings."""
        raw_live_setting = await self.http.get_live_setting(token=self.access_token)
        return raw_live_setting.content

    @overload
    async def set_live_setting(self, instance: BrodecastSetting) -> None:
        """Set the live settings.

        Parameters
        ----------
        instance : BrodecastSetting
            A instance of live setting.
        """
        ...

    @overload
    async def set_live_setting(
        self,
        title: Optional[str] = None,
        category: Optional[Category] = None,
        tags: Optional[list[str]] = None,
    ) -> None:
        """Set the live settings.

        Parameters
        ----------
        title : Optional[str]
            The name of brodecast.
        category : Optional[Category]
            The category of brodecast.
        tags : Optional[list[str]]
            The tags of brodecast.
        """
        ...

    @refreshable
    async def set_live_setting(
        self,
        instance: BrodecastSetting = None,
        title: Optional[str] = None,
        category: Optional[Category] = None,
        tags: Optional[list[str]] = None,
    ) -> None:
        if instance is not None:
            await self.http.set_live_setting(
                token=self.access_token,
                default_live_title=instance.title,
                category_id=instance.category.id,
                category_type=instance.category.type,
                tags=instance.tags,
            )
            return
        await self.http.set_live_setting(
            token=self.access_token,
            default_live_title=title,
            category_id=category.id if category is not None else None,
            category_type=category.type if category is not None else None,
            tags=tags,
        )
        return

    @refreshable
    async def get_stream_key(self) -> str:
        """Get the stream key to brodecast"""
        stream_key = await self.http.get_stream_key(token=self.access_token)
        return stream_key.content["streamKey"]

    @refreshable
    async def add_restrict_channel(self, user_id: str) -> None:
        """Add an user to restrict activity.

        Parameters
        ----------
        user_id : str
            A channel id of user to add restrict activity.
        """
        await self.http.add_restrcit_user(
            token=self.access_token, target_channel_id=user_id
        )
        return

    @refreshable
    async def remove_restrict_channel(self, user_id: str) -> None:
        """Remove an user to restrict activity.

        Parameters
        ----------
        user_id : str
            A channel id of user to remove restrict activity.
        """
        await self.http.remove_restrcit_user(
            token=self.access_token, target_channel_id=user_id
        )
        return

    @refreshable
    async def get_restriction(self, size: int = 20) -> SearchResult[RestrictUser]:
        """Get users with restricted activities.

        Parameters
        ----------
        size : Optional[int], optional
            A number of lives to load at once, by default 20
        """
        result = await self.http.get_restrcit_users(token=self.access_token, size=size)
        data = result.content
        data._next_method = self.http.get_restrcit_users
        data._next_method_key_argument = {"size": size}
        return data

    @refreshable
    async def get_channel_administrator(self) -> list[ChannelPermission]:
        """Get users who have channel management permission."""
        result = await self.http.get_channel_administrator(token=self.access_token)
        return result.content

    @refreshable
    async def get_followers(
        self, page: int = 0, size: int = 30
    ) -> ChannelSearchResult[FollowerInfo]:
        """Get followers for this channel.

        Parameters
        ----------
        size : Optional[int]
            A number of followers to load at once, by default 30
        page : Optional[int]
            A page number to load at once, by default 0
        """
        result = await self.http.get_channel_followers(
            token=self.access_token, size=size, page=page
        )
        data = result.content
        data._next_method = self.http.get_channel_followers
        data._next_method_key_argument = {"size": size, "token": self.access_token}
        return result.content

    @refreshable
    async def get_subscribers(
        self,
        page: int = 0,
        size: int = 30,
        sort: Literal["RECENT", "LONGER"] = "RECENT",
    ) -> ChannelSearchResult[SubscriberInfo]:
        """Get subscribers for this channel.

        Parameters
        ----------
        size : Optional[int]
            A number of subscribers to load at once, by default 30
        sort : Optional[Literal['RECENT', 'LONGER']]
            A method of sorting subscribers.
        page : Optional[int]
            A page number to load at once, by default 0
        """
        result = await self.http.get_channel_subscribers(
            token=self.access_token, size=size, sort=sort, page=page
        )
        data = result.content
        data._next_method = self.http.get_channel_subscribers
        data._next_method_key_argument = {
            "size": size,
            "sort": sort,
            "token": self.access_token,
        }
        return result.content



"""MIT License

Copyright (c) 2024-2025 gunyu1019

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import asyncio
import inspect
import json
from typing import Callable, Any, TYPE_CHECKING, Optional

from .enums import EnginePacketType, SocketPacketType
from .message import Donation, Subscription, Message
from .session import EventSubscribeMessage

if TYPE_CHECKING:
    from .authorization import AccessToken
    from .http import ChzzkOpenAPISession


class ConnectionState:
    def __init__(
        self,
        dispatch: Callable[..., Any],
        handler: dict[str, Callable[..., Any]],
        http: ChzzkOpenAPISession,
        access_token: Optional[AccessToken] = None,
        variable_access_token: Optional[Callable[..., AccessToken]] = None,
        json_serializer: Optional[Callable[..., dict[str, Any]]] = None,
        debug_mode: bool = False,
    ):
        self.dispatch = dispatch
        self.handler = handler
        self.http = http

        self.gateway_parsers: dict[
            SocketPacketType | EnginePacketType, Callable[..., Any]
        ] = dict()
        self.event_parsers: dict[str, Callable[..., Any]] = dict()

        self.gateway_id: Optional[str] = None

        self.debug_mode = debug_mode
        self.access_token = access_token

        self.variable_access_token = variable_access_token or self.__dummy_method
        self.json_serializer = json_serializer or json.loads

        for _, func in inspect.getmembers(self):
            if hasattr(func, "__gateway_parsing__") and func.__gateway_parsing__:
                self.gateway_parsers[
                    func.__parsing_socket_packet__ or func.__parsing_engine_packet__
                ] = func

        for _, func in inspect.getmembers(self):
            if hasattr(func, "__event_parsing__"):
                self.event_parsers[func.__event_parsing__] = func

    @staticmethod
    def __dummy_method(payload):
        return payload

    @staticmethod
    def gateway_parsable(
        engine_packet_type: EnginePacketType,
        socket_packet_type: Optional[SocketPacketType] = None,
    ):
        def decorator(func: Callable[..., Any]):
            func.__gateway_parsing__ = True
            func.__parsing_engine_packet__ = engine_packet_type
            func.__parsing_socket_packet__ = socket_packet_type
            return func

        return decorator

    @staticmethod
    def event_parsable(event_name: str):
        def decorator(func: Callable[..., Any]):
            func.__event_parsing__ = event_name
            return func

        return decorator

    async def call_handler(self, key: str, *args: Any, **kwargs: Any):
        if key not in self.handler:
            return
        func = self.handler[key]

        if asyncio.iscoroutinefunction(func):
            await func(*args, **kwargs)
        else:
            func(*args, **kwargs)

    @gateway_parsable(EnginePacketType.MESSAGE, SocketPacketType.EVENT)
    async def _handle_evnet(self, data: list[Any]):
        event = str(data[0]).lower()
        arguments = data[1:]

        event_func = self.event_parsers.get(event)
        if event_func is not None:
            await event_func(*arguments)

        if not self.debug_mode:
            return
        self.dispatch("socket_event", event, *data)
        return

    @gateway_parsable(EnginePacketType.OPEN)
    async def _handle_eio_connect(self, open_packet: dict[str, Any]):
        sid = open_packet["sid"]
        self.gateway_id = sid
        if not self.debug_mode:
            return
        self.dispatch("engine_connect", sid)
        return

    @gateway_parsable(EnginePacketType.MESSAGE, SocketPacketType.CONNECT)
    async def _handle_connect(self, _):
        if not self.debug_mode:
            return
        self.dispatch("socket_connect")
        return

    @gateway_parsable(EnginePacketType.MESSAGE, SocketPacketType.DISCONNECT)
    async def _handle_disconnect(self, _):
        if not self.debug_mode:
            return
        self.dispatch("socket_disconnect")
        return

    @event_parsable("system")
    async def _handle_system(self, raw_data):
        data = self.json_serializer(raw_data)
        event_type = data["type"]
        event_data = data["data"]

        if event_type == "connected":
            session_id = event_data["sessionKey"]
            self.dispatch("connect", session_id)
            await self.call_handler("connect", session_id)
        elif event_type == "subscribed":
            event_message = EventSubscribeMessage.model_validate(event_data)
            self.dispatch("permission_invoked", event_message)
            await self.call_handler("channel_id_invoked", event_message.channel_id)
        elif event_type == "unsubscribed":
            event_message = EventSubscribeMessage.model_validate(event_data)
            self.dispatch("permission_reinvoked", event_message)
            await self.call_handler("channel_id_invoked", event_message.channel_id)
        elif event_type == "revoked":
            event_message = EventSubscribeMessage.model_validate(event_data)
            self.dispatch("permission_reinvoked_force", event_message)
            await self.call_handler("channel_id_invoked", event_message.channel_id)
        return

    @event_parsable("chat")
    async def _handle_chat(self, raw_data):
        data = self.json_serializer(raw_data)
        message = Message.model_validate(data)
        message._state = self
        message._access_token = self.access_token or self.variable_access_token(
            message.channel
        )
        self.dispatch("chat", message)
        return

    @event_parsable("donation")
    async def _handle_donation(self, raw_data):
        data = self.json_serializer(raw_data)
        message = Donation.model_validate(data)
        message._state = self
        message._access_token = self.access_token or self.variable_access_token(
            message.channel
        )
        self.dispatch("donation", message)
        return

    @event_parsable("subscription")
    async def _handle_subscription(self, raw_data):
        data = self.json_serializer(raw_data)
        message = Subscription.model_validate(data)
        message._state = self
        message._access_token = self.access_token or self.variable_access_token(
            message.channel
        )
        self.dispatch("subscription", message)
        return



"""MIT License

Copyright (c) 2024-2025 gunyu1019

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import datetime

from typing import Optional
from pydantic import Field, PrivateAttr, computed_field

from .base_model import ChzzkModel
from .category import CATEGORY_TYPE, Category
from .channel import Channel


class Live(ChzzkModel):
    live_id: int
    live_title: str
    live_image_url: str = Field(alias="liveThumbnailImageUrl")
    concurrent_user_count: int
    open_date: datetime.datetime
    adult: bool
    tags: list[str]

    _category_id: Optional[str] = PrivateAttr(default=None)
    _category_name: Optional[str] = PrivateAttr(default=None)
    _category_type: Optional[CATEGORY_TYPE] = PrivateAttr(default=None)

    _channel_id: Optional[str] = PrivateAttr(default=None)
    _channel_name: Optional[str] = PrivateAttr(default=None)
    _channel_image: Optional[str] = PrivateAttr(default=None)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._category_id = kwargs.pop("liveCategory")
        self._category_name = kwargs.pop("liveCategoryValue")
        self._category_type = kwargs.pop("categoryType")

        self._channel_id = kwargs.pop("channelId")
        self._channel_name = kwargs.pop("channelName")
        self._channel_image = kwargs.pop("channelImageUrl")

    @computed_field
    @property
    def category(self) -> Optional[Category]:
        if self._category_id is None:
            return
        return Category(
            categoryId=self._category_id,
            categoryValue=self._category_name,
            categoryType=self._category_type,
        )

    @computed_field
    @property
    def channel(self) -> Channel:
        return Channel(
            channelId=self._channel_id,
            channelName=self._channel_name,
            channelImageUrl=self._channel_image,
        )


class BrodecastSetting(ChzzkModel):
    title: str = Field(alias="defaultLiveTitle")
    category: Category
    tags: list[str]



"""MIT License

Copyright (c) 2024-2025 gunyu1019

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

from functools import wraps
from typing import Generic, TypeVar, Optional, TYPE_CHECKING

from pydantic import BaseModel, ConfigDict, Extra, PrivateAttr
from pydantic.alias_generators import to_camel

from ..base_model import ChzzkModel
from ..base_model import Content

if TYPE_CHECKING:
    from .manage.manage_client import ManageClient

T = TypeVar("T")


class ManagerClientAccessable(BaseModel):
    _manage_client: Optional[ManageClient] = PrivateAttr(default=None)

    @staticmethod
    def based_manage_client(func):
        @wraps(func)
        async def wrapper(self, *args, **kwargs):
            if not self.is_interactable:
                raise RuntimeError(
                    f"This {self.__class__.__name__} is intended to store data only."
                )
            return await func(self, *args, **kwargs)

        return wrapper

    @property
    def channel_id(self) -> str:
        if not self.is_interactable:
            raise RuntimeError(
                f"This {self.__class__.__name__} is intended to store data only."
            )
        return self._manage_client.channel_id

    @property
    def is_interactable(self):
        """Ensure this model has access to interact with manage client."""
        return self._manage_client is not None

    def _set_manage_client(self, client: ManageClient):
        if self.is_interactable:
            raise ValueError("Manage Client already set.")
        self._manage_client = client
        return self



"""MIT License

Copyright (c) 2024-2025 gunyu1019

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import asyncio
import aiohttp
import logging
import time

from typing import Literal, TYPE_CHECKING
from yarl import URL

from .base_model import ChzzkModel
from .enums import EnginePacketType, SocketPacketType
from .error import HTTPException, ChatConnectFailed, ReceiveErrorPacket
from .packet import Packet
from .payload import Payload

if TYPE_CHECKING:
    from typing import Any, Callable, Optional
    from .state import ConnectionState

_log = logging.getLogger(__name__)


class OpenPacketInfo(ChzzkModel):
    sid: str
    upgrades: list[str]
    ping_interval: int
    ping_timeout: int


class ChzzkGateway:
    def __init__(
        self,
        loop: asyncio.AbstractEventLoop,
        base_url: URL,
        session: aiohttp.ClientSession,
        current_transport: Literal["polling", "websocket"],
        open_packet_info: OpenPacketInfo,
        session_id: Optional[str] = None,
        event_hook=None,
    ):
        self.current_transport = current_transport
        self.upgrades = open_packet_info.upgrades
        self.ping_interval = open_packet_info.ping_interval / 1000.0
        self.ping_timeout = open_packet_info.ping_timeout / 1000.0

        self.base_url = base_url
        self.session_id = session_id or open_packet_info.sid

        self.loop = loop
        self.session: aiohttp.ClientSession = session
        self.websocket: Optional[aiohttp.ClientWebSocketResponse] = None

        self.is_connected = True

        self._heartbeat_receive_event = asyncio.Event()
        self._heartbeat_receive_event.clear()

        if self.current_transport == "websocket":
            self._write = self._write_websocket
            self._read_loop = self._read_websocket
        else:
            self._write = self._write_polling
            self._read_loop = self._read_polling

        self._read_background_loop: Optional[asyncio.Task] = None

        _log.debug(f"Success connected to {self.base_url.host} with socket.io gateway")

        self._event_hook: dict[
            SocketPacketType | EnginePacketType, Optional[Callable[..., Any]]
        ] = event_hook or {
            key: None for key in list(SocketPacketType) + list(EnginePacketType)
        }

        _log.debug("Start handshake task")
        self._ping_loop_task = loop.create_task(self._ping_loop())

    def set_hook(
        self, event: SocketPacketType | EnginePacketType, coro_func: Callable[..., Any]
    ):
        self._event_hook[event] = coro_func

    def remove_hook(self, event: SocketPacketType | EnginePacketType):
        self._event_hook[event] = None

    @staticmethod
    def _get_engineio_url(
        url: str | URL,
        engine_path: str,
        transport: Literal["polling", "websocket"],
        ssl: bool = True,
    ) -> URL:
        new_url = url
        if not isinstance(url, URL):
            new_url = URL(url)

        query = new_url.query.copy()

        if transport == "polling":
            new_url = new_url.with_scheme("https" if ssl else "http")
        else:
            new_url = new_url.with_scheme("wss" if ssl else "ws")

        new_url = new_url.with_path(f"{engine_path}/")

        query.extend(**{"EIO": "3", "transport": transport})
        new_url = new_url.with_query(query)
        return new_url

    @staticmethod
    def _get_timestamp_url(url: URL) -> URL:
        query = url.query.copy()
        if "t" in query.keys():
            return url

        query["t"] = str(time.time())
        return url.with_query(query)

    @classmethod
    async def connect(
        cls,
        url: str | URL,
        state: ConnectionState,
        loop: asyncio.AbstractEventLoop,
        session: aiohttp.ClientSession,
        engine_path: Optional[str] = None,
    ):
        engine_path = engine_path or "socket.io"
        gateway = await cls._connect_polling(url, engine_path, loop, session)

        for event, parsing_func in state.gateway_parsers.items():
            if parsing_func is None:
                continue
            gateway.set_hook(event, parsing_func)

        return gateway

    @classmethod
    async def _connect_polling(
        cls,
        url: str | URL,
        engine_path: str,
        loop: asyncio.AbstractEventLoop,
        session: aiohttp.ClientSession,
        event_hook: Optional[
            dict[EnginePacketType | SocketPacketType, Callable[..., Any]]
        ] = None,
    ):
        base_url = cls._get_engineio_url(
            url=url, engine_path=engine_path, transport="polling"
        )
        base_url = cls._get_timestamp_url(base_url)
        connection_response = await session.request("GET", base_url)

        if connection_response.status < 200 or connection_response.status >= 300:
            raise ChatConnectFailed.polling_connect_failed(connection_response.status)

        raw_payload = await connection_response.read()
        payload = Payload.decode(raw_payload)

        raw_open_packet = payload.packets[0]
        open_packet = OpenPacketInfo.model_validate(raw_open_packet.data)

        if "websocket" in open_packet.upgrades:
            try:
                new_cls = await cls._connect_websocket(
                    url=url,
                    engine_path=engine_path,
                    loop=loop,
                    session=session,
                    open_packet=open_packet,
                    event_hook=event_hook,
                )
            except (
                aiohttp.client_exceptions.WSServerHandshakeError,
                aiohttp.client_exceptions.ServerConnectionError,
                aiohttp.client_exceptions.ClientConnectionError,
                ChatConnectFailed,
            ):
                _log.info(
                    "Failed upgrade to websocket transport: use polling transport."
                )
                pass
            else:
                for packet in payload.packets[1:]:
                    new_cls.received_message(packet)
                return new_cls

        query = base_url.query.copy()
        query["sid"] = open_packet.sid
        base_url = base_url.with_query(query)

        new_cls = cls(
            loop=loop,
            base_url=base_url,
            session=session,
            current_transport="polling",
            open_packet_info=open_packet,
            session_id=open_packet.sid,
            event_hook=event_hook,
        )
        for packet in payload.packets:
            new_cls.received_message(packet)
        return new_cls

    @classmethod
    async def _connect_websocket(
        cls,
        url: str | URL,
        engine_path: str,
        loop: asyncio.AbstractEventLoop,
        session: aiohttp.ClientSession,
        open_packet: Optional[OpenPacketInfo] = None,  # For update,
        event_hook: Optional[
            dict[EnginePacketType | SocketPacketType, Callable[..., Any]]
        ] = None,
    ):
        base_url = cls._get_engineio_url(
            url=url, engine_path=engine_path, transport="websocket"
        )
        base_url = cls._get_timestamp_url(base_url)

        if open_packet is not None:
            query = base_url.query.copy()
            query["sid"] = open_packet.sid
            base_url = base_url.with_query(query)
            upgrade = True
        else:
            upgrade = False

        websocket = await session.ws_connect(base_url)

        if upgrade:
            ping_packet = Packet(EnginePacketType.PING, data="probe")

            await websocket.send_str(ping_packet.encode())
            raw_pong_packet = (await websocket.receive()).data
            pong_packet = Packet.decode(raw_pong_packet)

            if (
                pong_packet.engine_packet_type != EnginePacketType.PONG
                or pong_packet.data != "probe"
            ):
                raise ChatConnectFailed.websocket_upgrade_failed()

            upgrade_packet = Packet(EnginePacketType.UPGRADE)
            await websocket.send_str(upgrade_packet.encode())
        else:
            raw_open_packet = (await websocket.receive()).data
            raw_open_packet = Packet.decode(raw_open_packet)
            open_packet = OpenPacketInfo.model_validate(raw_open_packet.data)

            query = base_url.query.copy()
            query["sid"] = open_packet.sid
            base_url = base_url.with_query(query)

        session_id = open_packet.sid

        new_cls = cls(
            loop=loop,
            base_url=base_url,
            session=session,
            current_transport="websocket",
            open_packet_info=open_packet,
            session_id=session_id,
            event_hook=event_hook,
        )
        await new_cls.received_message(
            Packet(
                engine_packet_type=EnginePacketType.OPEN, data=open_packet.model_dump()
            )
        )
        new_cls.websocket = websocket
        return new_cls

    async def _read_polling(self):
        base_url = self._get_timestamp_url(self.base_url)
        response = await self.session.request(
            "GET", base_url, timeout=max(self.ping_interval, self.ping_timeout) + 5
        )

        if response.status >= 300 and response.status < 200:
            raise ReceiveErrorPacket(self.current_transport, self.status)

        raw_payload = await response.read()
        payload = Payload.decode(raw_payload)

        for packet in payload.packets:
            await self.received_message(packet)

    async def _read_websocket(self):
        message = await self.websocket.receive(
            timeout=self.ping_interval + self.ping_timeout
        )
        if message.type is aiohttp.WSMsgType.TEXT:
            data = message.data
            packet = Packet.decode(data)
            await self.received_message(packet)
        elif message.type == aiohttp.WSMsgType.ERROR:
            raise ReceiveErrorPacket(self.current_transport, self.data)

    async def _write_polling(self, data: Packet):
        write_response = await self.session.request(
            "POST", self.base_url, data=data.encode()
        )
        if write_response.status < 200 and write_response.status >= 300:
            raise HTTPException(write_response.status)
        return

    async def _write_websocket(self, data: Packet):
        await self.websocket.send_bytes(data.encode())
        return

    async def _ping_loop(self):
        while self.is_connected:
            _log.debug("Send Ping packet to server for heartbeat")
            await self.send_ping()
            try:
                await asyncio.wait_for(
                    self._heartbeat_receive_event.wait(), timeout=self.ping_timeout
                )
            except (asyncio.Timeout, asyncio.CancelledError):
                raise ConnectionError("PONG response has not been received.")
            _log.debug("Received Pong packet from server.")
            await asyncio.sleep(self.ping_interval)

    async def read(self):
        while self.is_connected:
            await self._read_loop()

    def read_in_background(self) -> asyncio.Task:
        task = self.loop.create_task(self.read())
        self._read_background_loop = task
        return task

    async def received_message(self, data: Packet):
        _log.debug(
            f"Received Packet (Engine Packet Type: {data.engine_packet_type}, "
            f"Socket Packet Type: {data.socket_packet_type}, Data: {data.data})"
        )
        if data.is_socket_packet:
            if (
                data.socket_packet_type == SocketPacketType.EVENT
                and data.id is not None
            ):
                await self.send_ack(data.id)

            func = self._event_hook.get(data.socket_packet_type)
            if func is not None:
                await func(data.data)
            return

        if data.engine_packet_type == EnginePacketType.PONG:
            self._heartbeat_receive_event.set()
        elif data.engine_packet_type == EnginePacketType.CLOSE:
            _log.warning("Received close packet, close connection.")
            await self.disconnect()

        func = self._event_hook.get(data.engine_packet_type)
        if func is not None:
            await func(data.data)
        return

    async def send(self, packet: Payload | Packet):
        if isinstance(packet, Packet):
            packet = Payload(packets=[packet])
        await self._write(packet)

    async def disconnect(self):
        if not self.is_connected:
            return

        self._ping_loop_task.cancel()
        await self.send_disconnet()
        self.is_connected = False

        if (
            self._read_background_loop is not None
            and not self._read_background_loop.cancelled()
        ):
            self._read_background_loop.cancel()
        await self.websocket.close()

    async def send_ping(self, message: Optional[str] = None):
        await self.send(Packet(EnginePacketType.PING, data=message))

    async def send_disconnet(self):
        await self.send(Packet(EnginePacketType.MESSAGE, SocketPacketType.DISCONNECT))
        await self.send(Packet(EnginePacketType.CLOSE))

    async def send_ack(self, pakcet_id: int):
        await self.send(
            Packet(EnginePacketType.MESSAGE, SocketPacketType.ACK, packet_id=pakcet_id)
        )



"""MIT License

Copyright (c) 2024-2025 gunyu1019

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

from typing import Generic, Optional, TypeVar, TYPE_CHECKING
from pydantic import BaseModel, ConfigDict, Extra, PrivateAttr
from pydantic.alias_generators import to_camel

if TYPE_CHECKING:
    from typing import Any, Callable, Coroutine, Optional

T = TypeVar("T")


class ChzzkModel(BaseModel):
    model_config = ConfigDict(
        alias_generator=to_camel, frozen=True, extra=Extra.allow  # prevent exception.
    )

    @staticmethod
    def special_date_parsing_validator(value: T) -> T:
        if not isinstance(value, str):
            return value
        return value.replace("+09", "")


class Content(ChzzkModel, Generic[T]):
    code: int
    message: Optional[str]
    content: Optional[T]


class SearchResult(ChzzkModel, Generic[T]):
    data: list[T]

    _page: Optional[dict[str] | int] = PrivateAttr(default=None)
    _next_method: Optional[Callable[..., Coroutine[Any, Any, Content[T]]]] = (
        PrivateAttr(default=None)
    )
    _next_method_arguments: Optional[tuple[Any]] = PrivateAttr(default=None)
    _next_method_key_argument: Optional[dict[str, Any]] = PrivateAttr(default=None)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if "page" in kwargs.keys():
            self._page = kwargs.pop("page")

    async def _next(self, *args, **kwargs) -> Optional[SearchResult]:
        if self._next_method is None:
            return None

        next_method_arguments = self._next_method_arguments or tuple()
        next_method_key_argument = self._next_method_key_argument or dict()
        next_method_arguments += args
        next_method_key_argument.update(kwargs)

        result = await self._next_method(
            *next_method_arguments, **next_method_key_argument
        )
        data = result.content
        data._next_method = self._next_method
        data._next_method_arguments = self._next_method_arguments
        data._next_method_key_argument = self._next_method_key_argument
        return data

    async def next(self) -> Optional[SearchResult]:
        if self._page is None or self._next_method is None:
            raise RuntimeError(f"This search result has only one result.")
        next_id = self._page["next"]
        result = await self._next(next=next_id)
        return result

    def __getitem__(self, index):
        if isinstance(index, (slice, int)):
            return self.data[index]
        raise TypeError("'SearchResult' object is not subscriptable.")

    def __len__(self) -> int:
        return len(self.data)


class ChannelSearchResult(SearchResult, Generic[T]):
    total_count: int
    total_pages: int

    def __init__(self, *args, **kwargs):
        print(kwargs)
        super().__init__(*args, **kwargs)

    async def next(self) -> Optional[ChannelSearchResult]:
        if self._page + 1 >= self.total_pages:
            raise TypeError("This is the last page of search result.")
        page_id = self._page + 1
        result = await self._next(page=page_id)
        return result

    async def previous(self) -> Optional[ChannelSearchResult]:
        if self._page - 1 < 0:
            raise TypeError("This is the first page of search result.")
        page_id = self._page - 1
        result = await self._next(page=page_id)
        return result
